BASE_MODEL = "base_model"
INPUT_DIM = "input_dim"
ADJ = "adjacency_matrix"
ADJ_MOD = "adjacency_modifier"

OPT_TYPE = "opt_type"
OPT_ARGS = "opt_args"

FLOW_STEPS = "flow_steps"
FLOW_PERMUTE = "flow_permute"

ODENET_HID = "odenet_hid_dim"
ODENET_LIN = "odenet_lin_type"
ODENET_ACT = "odenet_act_type"

ODENET_DIV_FN = "ode_divergence_fn"
ODE_TIME = "ode_time_length"
ODE_TRAIN_T = "ode_train_T"
ODE_SOLVER_TYPE = "ode_solver_type"
ODE_SOLVER_ATOL = "ode_solver_atol"
ODE_SOLVER_RTOL = "ode_solver_rtol"
ODE_SOLVER_STEP = "ode_solver_stepsize"

COND_TYPE = "conditioner_type"
COND_HID = "conditioner_hid_dim"
COND_ACT = "conditioner_act_type"

N_PARAM_PER_VAR = "n_param_per_var"

GNF_HOT = "gnf_conditioner_hot_encoding"

NORM_TYPE = "normalizer_type"
UMNN_INT_HID = "umnn_int_hid_dim"
UMNN_INT_SOLVER = "umnn_int_solver"
UMNN_INT_STEP = "umnn_int_step"
